@@ -18,8 +18,8 @@ ChernoCommandType = Union[Command[ChernoActor], FakeCommand]
 
 
 from .acquire import *
+from .converge import *
 from .guide import *
-from .mock import *
 from .offset import *
 from .scale import *
 from .set import *
