@@ -8,6 +8,7 @@
 
 from __future__ import annotations
 
+import asyncio
 from functools import partial
 
 from typing import TYPE_CHECKING
@@ -66,6 +67,17 @@ __all__ = ["acquire"]
     is_flag=True,
     help="Applies the full correction once. Cannot be used with --continuous.",
 )
+@click.option(
+    "-w",
+    "--wait",
+    type=float,
+    help="Time to wait between iterations.",
+)
+@click.option(
+    "--no-block",
+    is_flag=True,
+    help="Returns immediatly after starting the acquisition loop.",
+)
 async def acquire(
     command: ChernoCommandType,
     exposure_time: float | None = None,
@@ -74,6 +86,8 @@ async def acquire(
     apply: bool = True,
     full: bool = False,
     cameras: str | None = None,
+    wait: float | None = None,
+    no_block: bool = False,
 ):
     """Runs the acquisition procedure."""
 
@@ -82,6 +96,11 @@ async def acquire(
     if count is not None and continuous is True:
         return command.fail("--count and --continuous are incompatible.")
 
+    _exposure_loop = command.actor.state._exposure_loop
+    if _exposure_loop is not None and not _exposure_loop.done():
+        command.warning("An active Exposer loop was found. Cancelling it.")
+        _exposure_loop.cancel()
+
     count = count or 1
 
     if cameras is None:
@@ -96,23 +115,32 @@ async def acquire(
             command.actor.state.exposure_time = exposure_time
 
     acquisition = Acquisition(config["observatory"])
+    command.actor.state._acquisition_obj = acquisition  # To update PID coeffs.
 
     callback = partial(
         acquisition.process,
         correct=apply,
         full_correction=full,
-        scale_rms=True,
+        wait_for_correction=(wait is None),
     )
     exposer = Exposer(command, callback=callback)
 
-    try:
-        await exposer.loop(
+    command.actor.state._exposure_loop = asyncio.create_task(
+        exposer.loop(
             None,
             count=count if continuous is False else None,
             timeout=25,
             names=names,
+            delay=wait or 0.0,
         )
-    except ExposerError as err:
-        return command.fail(f"Acquisition failed: {err}")
+    )
+
+    if not no_block:
+        try:
+            await command.actor.state._exposure_loop
+        except ExposerError as err:
+            return command.fail(f"Acquisition failed: {err}")
+    else:
+        return command.finish("Finishing due to --no-block. The guide loop is running.")
 
     return command.finish()
