@@ -15,7 +15,7 @@ import warnings
 from copy import deepcopy
 from dataclasses import dataclass, field
 
-from typing import Any, TypeVar
+from typing import Any, TypeVar, cast
 
 import matplotlib.pyplot as plt
 import numpy
@@ -80,7 +80,8 @@ class Extraction:
         **params,
     ):
 
-        self.pixel_scale = pixel_scale or config["pixel_scale"][observatory.upper()]
+        self.observatory = observatory.upper()
+        self.pixel_scale = pixel_scale or config["pixel_scale"]
 
         self.params = deepcopy(config["extraction"])
         self.params.update(params)
@@ -132,14 +133,23 @@ class Extraction:
         regions["mjd"] = mjd
         regions["exposure"] = exp_no
         regions["camera"] = cam_no
+        regions["fwhm_valid"] = 1
 
         if len(regions) > 0 and self.params["rejection_method"] is not None:
             self.reject(regions)
 
-        # Prevent NaNs here since this is output to the headers.
-        fwhm_median = numpy.round(regions.loc[regions.valid == 1].fwhm.median(), 3)
-        if numpy.isnan(fwhm_median):
-            fwhm_median = -999.0
+        valid = regions.loc[regions.fwhm_valid == 1]
+
+        if len(valid) > 0:
+            perc_50 = numpy.percentile(valid.fwhm, 50)
+            fwhm_median = valid.loc[valid.fwhm < perc_50].fwhm.median()
+            fwhm_median_round = float(numpy.round(fwhm_median, 3))
+
+            # Prevent NaNs here since this is output to the headers.
+            if numpy.isnan(fwhm_median_round):
+                fwhm_median_round = -999.0
+        else:
+            fwhm_median_round = -999.0
 
         extraction_data = ExtractionData(
             str(image),
@@ -156,7 +166,7 @@ class Extraction:
             regions=regions,
             nregions=len(regions),
             nvalid=sum(regions.valid == 1),
-            fwhm_median=fwhm_median,
+            fwhm_median=fwhm_median_round,
             focus_offset=config["cameras"]["focus_offset"][camera],
         )
 
@@ -198,7 +208,10 @@ class Extraction:
         return output_dir / basename
 
     def _process_sextractor(
-        self, data: numpy.ndarray, path: pathlib.Path, plot: bool | None = None
+        self,
+        data: numpy.ndarray,
+        path: pathlib.Path,
+        plot: bool | None = None,
     ):
         """Process image data using SExtractor."""
 
@@ -214,7 +227,7 @@ class Extraction:
                 fig, ax = plt.subplots()
                 ax.imshow(back, origin="lower")
                 ax.set_title("Background: " + path.parts[-1])
-                ax.set_gid(False)
+                ax.grid(False)
                 fig.savefig(str(output_root) + "-background.pdf")
 
         regions = sep.extract(
@@ -239,8 +252,7 @@ class Extraction:
             regions = regions.loc[regions.valid == 1, :]
             regions = regions.head(self.params["max_stars"])
 
-        regions = calculate_fwhm_from_ellipse(regions)
-        regions.fwhm *= self.pixel_scale
+        regions = calculate_fwhm_from_ellipse(regions, self.pixel_scale)
 
         if plot:
             data_back = data - back.back()
@@ -391,6 +403,8 @@ class Extraction:
 
         for index, row in regions.iterrows():
 
+            index = cast(Any, index)
+
             # Ignore detections that we have already marked as invalid in SExtractor.
             if row.valid == 0:
                 continue
@@ -432,7 +446,7 @@ class Extraction:
 
                     gauss_init = Gaussian1D(
                         amplitude=data_marginal.max(),
-                        stddev=row.fwhm_sextractor / p_scale,
+                        stddev=row.fwhm_sextractor / p_scale / gaussian_sigma_to_fwhm,
                         mean=mid,
                     )
                     g = fitter(gauss_init, x_mesh, data_marginal)
@@ -448,7 +462,7 @@ class Extraction:
                     trap_init = Trapezoid1D(
                         amplitude=data_marginal.max(),
                         x_0=mid,
-                        width=row.fwhm_sextractor / p_scale,
+                        width=row.fwhm_sextractor / p_scale / gaussian_sigma_to_fwhm,
                         slope=0.1,
                     )
                     t = fitter(trap_init, x_mesh, data_marginal)
@@ -475,28 +489,32 @@ class Extraction:
             # plt.close("all")
 
             regions.loc[index, "gaussian_fit"] = int(gauss_valid)
-            regions.loc[index, ["x_gaussian", "y_gaussian"]] = gauss_centroids[::-1]
-            regions.loc[index, "fwhm_gaussian"] = numpy.mean(gauss_fwhm)
+            regions.loc[index, "x_gaussian"] = gauss_centroids[::-1][0]
+            regions.loc[index, "y_gaussian"] = gauss_centroids[::-1][1]
+            regions.loc[index, "fwhm_gaussian"] = float(numpy.mean(gauss_fwhm))
             gauss_residual_mean = (numpy.array(gauss_residual) ** 2).sum() ** 0.5
             regions.loc[index, "residual_gaussian"] = gauss_residual_mean
 
             regions.loc[index, "trapezoid_fit"] = int(trap_valid)
-            regions.loc[index, ["x_trapezoid", "y_trapezoid"]] = trap_centroids[::-1]
-            regions.loc[index, "fwhm_trapezoid"] = numpy.mean(trap_fwhm)
+            regions.loc[index, "x_trapezoid"] = trap_centroids[::-1][0]
+            regions.loc[index, "y_trapezoid"] = trap_centroids[::-1][1]
+            regions.loc[index, "fwhm_trapezoid"] = float(numpy.mean(trap_fwhm))
             trap_residual_mean = (numpy.array(trap_residual) ** 2).sum() ** 0.5
             regions.loc[index, "residual_trapezoid"] = trap_residual_mean
 
             if not gauss_valid and not trap_valid:
                 continue
             elif gauss_valid and gauss_residual_mean <= trap_residual_mean:
-                regions.loc[index, ["x_fit", "y_fit"]] = gauss_centroids[::-1]
-                regions.loc[index, ["fwhm"]] = numpy.mean(gauss_fwhm)
-                regions.loc[index, ["residual_fit"]] = gauss_residual_mean
+                regions.loc[index, "x_fit"] = gauss_centroids[::-1][0]
+                regions.loc[index, "y_fit"] = gauss_centroids[::-1][1]
+                regions.loc[index, "fwhm"] = float(numpy.mean(gauss_fwhm))
+                regions.loc[index, "residual_fit"] = gauss_residual_mean
                 regions.loc[index, "model_fit"] = "g"
             else:
-                regions.loc[index, ["x_fit", "y_fit"]] = trap_centroids[::-1]
-                regions.loc[index, ["fwhm"]] = numpy.mean(trap_fwhm)
-                regions.loc[index, ["residual_fit"]] = trap_residual_mean
+                regions.loc[index, "x_fit"] = trap_centroids[::-1][0]
+                regions.loc[index, "y_fit"] = trap_centroids[::-1][1]
+                regions.loc[index, "fwhm"] = float(numpy.mean(trap_fwhm))
+                regions.loc[index, "residual_fit"] = trap_residual_mean
                 regions.loc[index, "model_fit"] = "t"
 
         regions.loc[regions.model_fit == "", "valid"] = 0
@@ -524,12 +542,29 @@ class Extraction:
 
         method = self.params.get("rejection_method", "sigclip")
 
+        if self.star_finder in ["sextractor", "marginal"]:
+
+            # Filter out bad FWHM values.
+            ecc = numpy.sqrt(regions.a**2 - regions.b**2) / regions.a
+
+            filter = (
+                ((regions.a * self.pixel_scale) < 5)
+                & ((regions.a * self.pixel_scale) > 0.4)
+                & (regions.cpeak < 60000)
+                & (ecc < 0.7)
+            )
+
+            regions.loc[~filter, "fwhm_valid"] = 0
+
         if method == "sigclip":
-            fwhm = numpy.ma.array(regions.fwhm.values, mask=(regions.valid == 0))
+            fwhm = numpy.ma.array(
+                regions.fwhm.values,
+                mask=((regions.valid == 0) | (regions.fwhm_valid == 0)),
+            )
             sigma = self.params.get("reject_sigma", 3.0)
             sigma_clip = SigmaClip(sigma, cenfunc="median")
             masked: Any = sigma_clip(fwhm, masked=True)
-            regions.loc[masked.mask, "valid"] = 0
+            regions.loc[masked.mask, "fwhm_valid"] = 0
 
         elif method == "nreject":
             nreject = self.params.get("nreject", 3)
@@ -547,7 +582,7 @@ class Extraction:
                 else:
                     nreject = nreject // 2
 
-            regions.loc[~regions.index.isin(valid.index), "valid"] = 0
+            regions.loc[~regions.index.isin(valid.index), "fwhm_valid"] = 0
 
         else:
             raise ValueError(f"Invalid rejection method {method!r}.")
@@ -576,7 +611,7 @@ class Extraction:
             fig, ax = plt.subplots()
 
         ax.set_title(title or path.parts[-1])
-        ax.set_gid(False)
+        ax.grid(False)
 
         # Image
         ax.imshow(data, origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
@@ -617,6 +652,7 @@ class Extraction:
 
 def calculate_fwhm_from_ellipse(
     regions: pandas.DataFrame,
+    pixel_scale: float,
     a_col: str = "a",
     b_col: str = "b",
 ):
@@ -627,6 +663,8 @@ def calculate_fwhm_from_ellipse(
     regions
         The pandas data frame with the list of regions. Usually an output from
         ``sep``. Must include columns ``valid``, ``a``, and ``b``.
+    pixel_scale
+        The pixel scale in arcsec per pixel.
     a_col
         Column with the semi-major axis measurement.
     b_col
@@ -640,10 +678,8 @@ def calculate_fwhm_from_ellipse(
 
     """
 
-    fwhm = numpy.max(numpy.array([regions[a_col] * 2, regions[b_col] * 2]), axis=0)
-
-    # From Dylan
-    # fwhm = 2 * numpy.sqrt(numpy.log(2) * (regions[a_col] ** 2 + regions[b_col] ** 2))
+    fwhm_pixel = 2 * (numpy.log(2) * (regions[a_col] ** 2 + regions[b_col] ** 2)) ** 0.5
+    fwhm = pixel_scale * fwhm_pixel
 
     regions = regions.copy()
     regions["fwhm"] = fwhm
